/*
 * Copyright (c) 1996 Palevich-Chen Industries, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for all purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 * PALEVICH-CHEN INDUSTRIES MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT
 * THE SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE
 * FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

import java.io.*;

/**
 * Edict is a class that understands
 * <a href="http://www.rdt.monash.edu.au/~jwb/edict.html">Jim Breen's public domain
 * EDICT Japanese/English dictionary file.</a>
 * Requires the files edict and edict.jdx, available from
 * <a href="ftp://ftp.cc.monash.edu.au/pub/nihongo">
 * ftp://ftp.cc.monash.edu.au/pub/nihongo</a>.
 *
 * This class implements a dictionary that can be used to translate words between Japanese
 * and English.
 *
 * The Japanese text must be encoded in EUC, which is one of several non-Unicode
 * standards for encoding Japanese text. If you have text in EUC, you're all set. If you
 * have text in <a href="http://www.unicode.org">Unicode</a>, you must translate the text
 * into EUC before trying to look it
 * up in this dictionary. If you have text in Shift-JIS, you can use the utilities in
 * the CodeUtil class to convert the text to EUC. One good source of information on
 * converting between the various standards for Japanese text is
 * <a href="http://www.ora.de/gnn/bus/ora/item/ujip.html">
 * Understanding Japanese Information Processing</a>.
 *
 * Here are some examples of how to use this class:
 * <pre>
 * EDict edict = new EDict();
 * String entryForLetter = edict.entryAt("letter");
 * String entryForEUC_BBFA = edict.entryAt("\u00bb\00fa"); // Unicode \u5B57 - letter/character
 * </pre>
 *
 * Since there might be several entries for the same word (especially for the reverse
 * translation from English to Japanese) there is an iterator interface using the
 * class EDictIterator:
 *
 * <pre>
 * EDictIterator iterator = edict.entriesAt("letter");
 * while ( iterator.hasMoreElements() ){
 *  System.out.println((String) iterator.nextElement());
 * }
 * </pre>
 *
 * Since EUC is a 8-bit encoding, there are byte[] versions as well as String versions of
 * all the methods.
 *
 * @see		EDictIterator
 * @author  Jack Palevich
 */

public class EDict {

    /**
     * Constructs a new dictionary using the files edict and edict.jdx.
     */
    public EDict()
        throws IOException
    {
        this(new File("edict"));
    }

    /**
     * Constructs a new dictionary using the supplied file, and assumes that the
     * index is the same name, with ".jdx" appended.
     * @param dict the file that contains the dictionary.
     */
    public EDict(File dict)
        throws IOException
    {
        this(dict, new File(dict.getPath() + ".jdx"));
    }

    /**
     * Constructs a new dictionary using the supplied files.
     * @param dict the file that contains the dictionary.
     * @param index the file that contains the index.
     */
	public EDict(File dict, File index)
	    throws IOException
	{
	    this.dict = dict;
	    this.index = index;
	    // System.err.println("Reading dict..." + dict.getPath());
	    dictData = readAll(dict);

        if ( indexInRAM ){
        	if(autoCreateIndex){
        	    try {
        	        // System.err.println("Reading index..." + index);
        	        readIndex(index);
        	    } catch (IOException e) {
            	    createIndex(index);
        	    }
        	}
        	else {
        	    // System.err.println("Reading index..." + index);
        	    readIndex(index);
        	}
        }
        else {
            if ( ! index.exists() ) {
                if ( autoCreateIndex ) {
            	    createIndex(index);
            	}
            	else {
            	    throw new FileNotFoundException(index.toString());
            	}
            }
        }
	}

    /**
     * Returns the first entry for a given key. Can be either english or japanese.
     * @param key - the key, expressed as an EUC encoded String.
     * @return the first dictionary entry for that key.
     */

	public String entryAt(String key) {
	    return entryAt(toBytes(key));
	}

    /**
     * Returns the first entry for a given key. Can be either english or japanese.
     * @param key - the key, expressed as an EUC encoded byte array.
     */
	public String entryAt(byte[] key) {
	    return entryAt(indexOf(key));
	}

    /**
     * Returns an iterator for all the entries for a given key.
     * The key can be either english or japanese.
     * @param key - the key, expressed as an EUC encoded String.
     * @return - an EDictIterator that returns the entries that match the key.
     */
	public EDictIterator entriesAt(String key) {
	    return new EDictIterator(this, key);
	}

    /**
     * Returns an iterator for all the entries for a given key.
     * The key can be either english or japanese.
     * @param key - the key, expressed as an EUC encoded byte array.
     * @return - an EDictIterator that returns the entries that match the key.
     */
	public EDictIterator entriesAt(byte[] key) {
	    return new EDictIterator(this, key);
	}

    /**
     * Returns an iterator for all the entries in the dictionary.
     * Used for testing purposes.
     * @return - an EDictIterator that returns all the entries in the dictionary.
     */
	public EDictIterator allEntries() {
	    return new EDictIterator(this, "");
	}

    /**
     * Returns the entry for a given index of the dictionary. Used by the
     * EDictIterator class. You get an exception if you try < 0 or >= length()
     * @param index - the integer index 0..n of the total number of entries in the dictionary.
     * @return the associated entry.
     */
    public String entryAt(int index)
    {
        return getStringAt(getFullEntryIndex(getIndex(index)));
    }

    /**
     * Returns the number of entries in the dictionary.
     * @returns the number of entries in the dictionary.
     */
    public int length()
    {
        return indexData.length;
    }

    /**
     * Returns true of a key matches a particular index. Used by the
     * EDictIterator class.
     * @param key - the key, expressed as an EUC encoded string.
     * @param index - the integer index of the entry in the dictionary.
     * @return true if the key matches the entry for a given index.
     */
    public boolean keyMatches(String key, int index) {
        return keyMatches(toBytes(key), index);
    }

    /**
     * Returns true of a key matches a particular index. Used by the
     * EDictIterator class.
     * @param key - the key, expressed as an EUC encoded byte array.
     * @param index - the integer index of the entry in the dictionary.
     * @return true if the key matches the entry for a given index.
     */
    public boolean keyMatches(byte[] key, int index) {
        return kstrcmp(key, 0, dictData, getIndex(index)) == 0;
    }

    /**
     * Returns the index of the first entry that matches a particular key. Used by the
     * EDictIterator class.
     * @param key - the key, expressed as an EUC encoded string.
     * @returns the index if the first entry that matches a particular key.
     */

	public int indexOf(String key){
	    return indexOf(toBytes(key));
	}

    /**
     * Returns the index of the first entry that matches a particular key. Used by the
     * EDictIterator class.
     * @param key - the key, expressed as an EUC encoded byte array.
     * @returns the index if the first entry that matches a particular key.
     */

	public int indexOf(byte[] key){
	    // Binary search on index file
	    int numEntries = getIndexSize();
	    int low = 0;
	    int hi = numEntries;
	    int search = 0;
	    int searchCount = 40;
	    while ( low < hi - 1 ) {
	        search = (low + hi) / 2;
	        int entry = getIndex(search);
            if(debug){
                System.err.println("low: " + low + " hi: " + hi +
                    " search: " + search +
                    " key: " + new String(key,0) +
                    " entry: " + entry + "= " + getStringAt(entry));
            }
	        int result = kstrcmp(key, 0, dictData, entry);
	        if(debug){
	            System.err.println("result: " + result);
	        }
	        if ( result < 0 ) {
	            hi = search;
	        }
	        else if ( result == 0 ) {
	            break;
	        }
	        else {
	            low = search;
	        }
	        if ( searchCount-- < 0 ) {
	            // Search failed.
	            return 0;
	        }

	    }
	    // Search backward linearly for first key that matches
	    while ( search > 0 && kstrcmp(key, 0, dictData, getIndex(search)) <= 0 ) {
            if(debug) System.err.println("Backing up." + " search: " + search + " key: " + new String(key,0) + " entry: " + getIndex(search) + "= " + getEntry(search));
	        search--;
	    }
	    if ( search < 0 ) {
	        search = 0;
	    }
	    else {
	        search++;
	    }
        if(debug) System.err.println("Returning:" + search+"="+ getEntry(search));
	    return search;
	}

	static private byte[] toBytes(String s){
	    int length = s.length();
	    byte[] b = new byte[length];
	    s.getBytes(0, length, b, 0);
	    return b;
	}

	static private byte[] readAll(File file)
	    throws IOException
	{
	    byte[] data = null;
	    FileInputStream stream = null;
        stream = new FileInputStream(file);
        int size = stream.available();
        data = new byte[size];
        stream.read(data);
	    stream.close();
	    return data;
	}

	private void createIndex(File index)
	    throws IOException
	{
	    System.out.println("Creating index...");
	    indexData = createIndex();
	    System.out.println("Creating index file...");
        DataOutputStream indexStream = new DataOutputStream(
	        new BufferedOutputStream( new FileOutputStream(index)));
	    int indexLength = indexData.length;
	    indexStream.writeInt(dictData.length + jiver); // Magic cookie
        System.out.println("writing index data...");
	    for(int i = 0; i < indexLength; i++){
            int d = swapByteOrder(indexData[i]);
	        indexStream.writeInt(d);
	    }
	    indexStream.close();
        System.out.println("Index data written.");
	}

    // The index is stored in little-endian byte order, but Java likes big-endian.
    static final private int swapByteOrder(int d) {
        // Swap byte order
        int d2 = (d << 24) | ((d & 0xff00) << 8) | ((d & 0xff0000) >>> 8)
            | (d >>> 24);
        return d2;
    }

    private final int jiver = 15;

	private void readIndex(File index) throws IOException {
	    DataInputStream indexStream = new DataInputStream(
	        new BufferedInputStream( new FileInputStream(index)));
	    int cookie = swapByteOrder(indexStream.readInt()); // Magic cookie
	    int expectedCookie = dictData.length + jiver;
	    if ( cookie != expectedCookie ){
	        System.err.println("Bad cookie. Expected " + expectedCookie
	            + " got " + cookie);
	        // But try anyway. The new version might be upwards compatable.
	        // throw new IOException("Index does not match dictionary.");
	    }
	    int indexLength = (indexStream.available() + 3)/ 4 - 1;
	    indexData = new int[indexLength];
	    for(int i = 0; i < indexLength; i++){
	        int d = swapByteOrder(indexStream.readInt());
	        indexData[i] = d;
	    }
	    indexStream.close();
	}

	private int getIndexSize(){
	    if ( indexInRAM ){
	        return indexData.length;
	    }
	    else {
	        return (int) (index.length() + 3) / 4 - 1;
	    }
	}

	private int getIndex(int index){
	    int result = 0;
	    if ( indexInRAM ){
	        return result = indexData[index];
	    }
	    else {
	        try {
    	        DataInputStream stream = new DataInputStream(
    	            new FileInputStream(this.index));
    	        stream.skip(index * 4);
    	        result = swapByteOrder(stream.readInt());
    	        stream.close();
    	    } catch (IOException e) {
    	        throw new ArrayIndexOutOfBoundsException();
    	    }
	    }
    	return result-1;
    }

    private int getFullEntryIndex(int entryIndex){
        while ( entryIndex > 0 && dictData[entryIndex - 1] != '\n'){
            entryIndex--;
        }
        return entryIndex;
    }

    /* Partial entry */
	private String getEntry(int index){
        return getStringAt(getIndex(index));
    }

    private String getStringAt(int offset){
        for(int i = 0; i < recordSize; i++) {
            int c = dictData[offset + i];
            if ( c == '\n' ) {
                return new String(dictData, 0, offset, i);
            }
        }
        return new String(dictData, 0, offset, recordSize);
    }

    final int recordSize = 512;
    private int[] createIndex(){
        int indexSize = 500;
        int[] index = new int[indexSize];
        int usedIndex = 0;
        int dictLength = dictData.length;
        // record first entry
        index[usedIndex++] = 0;
        for(int i = 0; i < dictLength; i++ ) {
            // if ( i > 500000 ) break;
            int c = dictData[i];
            // Index the kanji.
            if ( c == '\n' || (c == '[' && i > 0 && dictData[i-1] == ' ')) {
                // System.out.println("i = " + i);
                if ( usedIndex >= indexSize ) {
                    int newIndexSize = 2 * indexSize;
                    int[] newIndex = new int[newIndexSize];
                    for ( int k = 0; k < indexSize; k++ ) { newIndex[k] = index[k]; }
                    indexSize = newIndexSize;
                    index = newIndex;
                }
                index[usedIndex++] = i + 1;
            }
        }

        // If the last character of the dictionary is a '\n', trim last entry
        if ( index[usedIndex-1] >= dictLength ) --usedIndex;

        int[] newIndex = new int[usedIndex];
        for ( int k = 0; k < usedIndex; k++ ) { newIndex[k] = index[k]; }
        indexSize = usedIndex;
        index = newIndex;

        sort(index, 0, indexSize-1);
        if ( !checkIndex(index) ) return null;
        return index;
    }

    private boolean checkIndex(int[] index) {
        boolean result = true;
        for ( int i = 1; i < index.length; i++ ){
            if ( kstrcmp(dictData, index[i-1], dictData, index[i]) > 0 ) {
                System.err.println("Out of order: " + (i-1));
                result = false;
            }
        }
        return result;
    }

    // ShellSort from http://www.dcc.uchile.cl/~oalonso/handbook/algs/4/413.sort.c.html

    private void sort(int[] r, int lo, int up){
        System.out.println("sort lo: " + lo + " up: " + up);
        for ( int d=up-lo+1; d>1; ) {
        // System.out.println("d:" + d);
           if (d<5)  d = 1;
              else   d = (5*d-1)/11;
        System.out.println(" d:" + d);
           /*** Do linear insertion sort in steps size d ***/
           for ( int i=up-d; i>=lo; i-- ) {
         // System.out.println(" i:" + i);
                int tempr = r[i];
                int j;
                for ( j=i+d; j<=up && (kstrcmp(dictData, tempr, dictData, r[j]) > 0); j+=d )
         // System.out.println(" j:" + j);
                     r[j-d] = r[j];
                r[j-d] = tempr;
                }
           }
    }

    /* If a < b returns < 0. If a == b returns 0, if a > b returns >0 */
    static private int kstrcmp(byte[] lhsData, int lhs, byte[] rhsData, int rhs)
    {
    	int i,c1 = 0,c2 = 0;
    /* effectively does a strnicmp on two "strings" within the dictionary,
       except it will make katakana and hirgana match (EUC A4 & A5) */
    	for (i = 0; i<20 ; i++)
    	{
    		boolean lhsend = ( c1 == ' ' || c1 == ']' );
    		boolean rhsend = ( c2 == ' ' || c2 == ']' );
    		try {
    		    c1 = 0xff & lhsData[lhs + i];
    		} catch (Throwable e) {
    		    lhsend = true;
    		}
    		try {
    		    c2 = 0xff & rhsData[rhs + i];
    		} catch (Throwable e) {
    		    rhsend = true;
    		}
    		if ( lhsend || rhsend ) return 0;

    		if ((i % 2) == 0)
    		{
    			if (c1 == 0xA5)
    			{
    				c1 = 0xA4;
    			}
    			if (c2 == 0xA5)
    			{
    				c2 = 0xA4;
    			}
    		}
    		if ((c1 >= 'A') && (c1 <= 'Z')) c1 |= 0x20;
    		if ((c2 >= 'A') && (c2 <= 'Z')) c2 |= 0x20;
    		if (c1 != c2 ) break;
    	}
    	return(c1-c2);
    }

	private boolean autoCreateIndex = false;
    private boolean debug = false;
    // It takes a long time to load the index into RAM - and it takes a
    // lot of RAM. It's faster to do file I/O as needed.
	private boolean indexInRAM = false;
	private File dict;
	private File index;
	private byte[] dictData;
	private int[] indexData;
}