/*
 * Copyright (c) 1996 Palevich-Chen Industries, Inc. All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for all purposes and without
 * fee is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 * PALEVICH-CHEN INDUSTRIES MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT
 * THE SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. SUN SHALL NOT BE LIABLE
 * FOR ANY DAMAGES SUFFERED BY LICENSEE AS A RESULT OF USING, MODIFYING
 * OR DISTRIBUTING THIS SOFTWARE OR ITS DERIVATIVES.
 */

import java.util.*;
import java.io.*;

/**
 * EDictIterator is a class that provides iterator protocol over an EDict.
 *
 * Here are some examples of how to use this class:
 * <pre>
 * EDictIterator iterator = edict.entriesAt("letter");
 * while ( iterator.hasMoreElements() ){
 *  System.out.println((String) iterator.nextElement());
 * }
 * </pre>
 *
 *
 * @see		EDict
 * @author  Jack Palevich
 */

class EDictIterator implements Enumeration {
    private EDict dict;
    private byte[] key;
    private int index;

    /**
     * Constructs a new iterator using the supplied dictionary.
     * @param dict the dictionary.
     * @param key the EUC-encoded string to look up.
     */
    public EDictIterator(EDict dict, String key) {
        this(dict, toBytes(key));
    }

    /**
     * Constructs a new iterator using the supplied dictionary.
     * @param dict the dictionary.
     * @param key the EUC-encoded byte array to look up.
     */
    public EDictIterator(EDict dict, byte[] key) {
        this.dict = dict;
        this.key = key;
        index = dict.indexOf(key);
    }

    public boolean hasMoreElements() {
        boolean result = false;
        try {
            result = dict.keyMatches(key, index);
        }
        catch ( ArrayIndexOutOfBoundsException e ) {
            result = false;
        }
        return result;
    }

    public Object nextElement() {
        if ( ! hasMoreElements() ) {
            throw new NoSuchElementException();
            }
        return dict.entryAt(index++);
    }

    private static byte[] toBytes(String key){
        int keyLength = key.length();
        byte[] keyBytes = new byte[keyLength];
        key.getBytes(0, keyLength, keyBytes,0);
        return keyBytes;
    }
}